#include <torch/extension.h>
#include <cuda.h>
#include <cuda_runtime.h>
#include <cmath>

#define MAX_OUT_FEATURES 4096
#define TILE_DIM 16
#define WARP_SIZE 32
#define BLOCK_SIZE 256

__constant__ float c_bias[MAX_OUT_FEATURES];
__constant__ float c_subtract[MAX_OUT_FEATURES];

//------------------------------------------------------------------------------
// Warp-level primitives for efficient reduction
__device__ __forceinline__ float warp_reduce_sum(float val) {
    #pragma unroll
    for (int offset = WARP_SIZE/2; offset > 0; offset /= 2) {
        val += __shfl_down_sync(0xffffffff, val, offset);
    }
    return val;
}

//------------------------------------------------------------------------------
// Vector load/store helpers for better memory throughput
struct Float4 {
    float4 data;
    
    __device__ __forceinline__ Float4() {}
    
    __device__ __forceinline__ void load(const float* ptr) {
        data = *reinterpret_cast<const float4*>(ptr);
    }
    
    __device__ __forceinline__ void store(float* ptr) {
        *reinterpret_cast<float4*>(ptr) = data;
    }
};

//------------------------------------------------------------------------------
// Optimized GEMM kernel with vectorized memory access
__global__ void vectorized_gemm_subtract_kernel(
    const float* __restrict__ x,
    const float* __restrict__ weight,
    float* __restrict__ out,
    const int batch_size,
    const int in_features,
    const int out_features
) {
    __shared__ float tile_x[TILE_DIM][TILE_DIM+1];
    __shared__ float tile_w[TILE_DIM][TILE_DIM+1];

    const int tx = threadIdx.x;
    const int ty = threadIdx.y;
    const int row = blockIdx.y * TILE_DIM + ty;
    const int col = blockIdx.x * TILE_DIM + tx;
    
    float sum = 0.0f;

    // Use vectorized loads where possible
    const int vec_size = 4;
    const int vec_in_features = in_features / vec_size;
    
    for (int t = 0; t < (in_features + TILE_DIM - 1) / TILE_DIM; t++) {
        if (row < batch_size && (t * TILE_DIM + tx) < in_features) {
            tile_x[ty][tx] = x[row * in_features + t * TILE_DIM + tx];
        }
        
        if (col < out_features && (t * TILE_DIM + ty) < in_features) {
            tile_w[ty][tx] = weight[col * in_features + t * TILE_DIM + ty];
        }
        
        __syncthreads();

        #pragma unroll
        for (int k = 0; k < TILE_DIM; k++) {
            sum += tile_x[ty][k] * tile_w[k][tx];
        }
        
        __syncthreads();
    }

    if (row < batch_size && col < out_features) {
        out[row * out_features + col] = sum + c_bias[col] - c_subtract[col];
    }
}

//------------------------------------------------------------------------------
// Fused kernel for reduction, activation, and residual add with minimal atomics
__global__ void fused_reduce_gelu_residual_kernel(
    const float* __restrict__ gemm_out,
    const float* __restrict__ original_x,
    float* __restrict__ out,
    const int batch_size,
    const int out_features,
    const int in_features
) {
    extern __shared__ float smem[];
    
    const int tid = threadIdx.x;
    const int bid = blockIdx.x;
    const int lane_id = tid % WARP_SIZE;
    const int warp_id = tid / WARP_SIZE;
    
    // First perform warp-level reduction
    float warp_sum = 0.0f;
    const int items_per_thread = (out_features + blockDim.x - 1) / blockDim.x;
    
    #pragma unroll
    for (int i = 0; i < items_per_thread; i++) {
        const int idx = tid + i * blockDim.x;
        if (idx < out_features) {
            warp_sum += gemm_out[bid * out_features + idx];
        }
    }
    
    // Warp-level reduction
    warp_sum = warp_reduce_sum(warp_sum);
    
    // Only first thread in each warp writes to shared memory
    if (lane_id == 0) {
        smem[warp_id] = warp_sum;
    }
    __syncthreads();
    
    // Final reduction and GELU in first warp
    if (warp_id == 0) {
        float sum = (tid < (blockDim.x / WARP_SIZE)) ? smem[tid] : 0.0f;
        sum = warp_reduce_sum(sum);
        
        if (lane_id == 0) {
            const float avg = sum / static_cast<float>(out_features);
            const float kAlpha = 0.044715f;
            const float kBeta = 0.7978845608f;
            const float inner = kBeta * (avg + kAlpha * avg * avg * avg);
            smem[0] = avg * 0.5f * (1.0f + tanhf(inner));
        }
    }
    __syncthreads();
    
    // Broadcast result and perform residual add with vectorized access
    const float gelu_val = smem[0];
    const int vec_elements = in_features / 4;
    
    // Vector loads/stores for residual addition
    for (int i = tid; i < vec_elements; i += blockDim.x) {
        Float4 orig, result;
        orig.load(original_x + bid * in_features + i * 4);
        
        result.data.x = orig.data.x + gelu_val;
        result.data.y = orig.data.y + gelu_val;
        result.data.z = orig.data.z + gelu_val;
        result.data.w = orig.data.w + gelu_val;
        
        result.store(out + bid * in_features + i * 4);
    }
    
    // Handle remaining elements
    for (int i = vec_elements * 4 + tid; i < in_features; i += blockDim.x) {
        out[bid * in_features + i] = original_x[bid * in_features + i] + gelu_val;
    }
}

torch::Tensor forward_cuda(
    const torch::Tensor& x,
    const torch::Tensor& weight,
    const torch::Tensor& bias,
    const torch::Tensor& subtract
) {
    const int batch_size = x.size(0);
    const int in_features = x.size(1);
    const int out_features = weight.size(0);
    
    auto x_ = x.contiguous();
    auto w_ = weight.contiguous();
    auto b_ = bias.contiguous();
    auto s_ = subtract.contiguous();
    
    cudaMemcpyToSymbol(c_bias, b_.data_ptr<float>(), out_features * sizeof(float));
    cudaMemcpyToSymbol(c_subtract, s_.data_ptr<float>(), out_features * sizeof(float));
    
    auto original_x = x_.clone();
    auto gemm_out = torch::empty({batch_size, out_features}, x.options());
    auto out = torch::empty({batch_size, in_features}, x.options());
    
    dim3 gemm_block(TILE_DIM, TILE_DIM);
    dim3 gemm_grid(
        (out_features + TILE_DIM - 1) / TILE_DIM,
        (batch_size + TILE_DIM - 1) / TILE_DIM
    );
    
    vectorized_gemm_subtract_kernel<<<gemm_grid, gemm_block>>>(
        x_.data_ptr<float>(),
        w_.data_ptr<float>(),
        gemm_out.data_ptr<float>(),
        batch_size,
        in_features,
        out_features
    );
    
    const int threads = BLOCK_SIZE;
    const int shared_mem_size = (threads / WARP_SIZE + 1) * sizeof(float);
    
    fused_reduce_gelu_residual_kernel<<<batch_size, threads, shared_mem_size>>>(
        gemm_out.data_ptr<float>(),
        original_x.data_ptr<float>(),
        out.data_ptr<float>(),
        batch_size,
        out_features,
        in_features
    );
    
    return out;
}

PYBIND11_MODULE(TORCH_EXTENSION_NAME, m) {
    m.def("forward", &forward_cuda, 
          "Optimized CUDA implementation with minimal atomics and vectorized memory access");
}